"""
originpro
A package for interacting with Origin software via Python.
Copyright (c) 2020 OriginLab Corporation
"""
#from pprint import pprint
from .config import po
from .base import *
from .utils import *


class Axis:
    """
    This class represents an instance of Axis on a GLayer.
    """

    def __init__(self, obj, ax):
        self.layer = obj
        types = {
            'x':2,'y':4,'z':1
            }
        self.type = types[ax]
        self.ax = ax

    @property
    def sfrom(self):
        """get scale limit from value"""
        return self.layer.GetNumProp('{0}.from'.format(self.ax))
    @sfrom.setter
    def sfrom(self, value):
        """set scale limit from value"""
        self.layer.SetNumProp('{0}.from'.format(self.ax), value)

    @property
    def sto(self):
        """get scale limit to value"""
        return self.layer.GetNumProp('{0}.to'.format(self.ax))
    @sto.setter
    def sto(self, value):
        """set scale limit to value"""
        self.layer.SetNumProp('{0}.to'.format(self.ax), value)

    @property
    def sstep(self):
        """get scale limit increment"""
        return self.layer.GetNumProp('{0}.inc'.format(self.ax))
    @sstep.setter
    def sstep(self, value):
        """set scale limit increment"""
        self.layer.SetNumProp('{0}.inc'.format(self.ax), value)

    @property
    def scale(self) -> int:
        """get axis scale type"""
        return int(self.layer.GetNumProp(f'{self.ax}.type'))

    @scale.setter
    def scale(self, value):
        """set axis scale type"""
        if isinstance(value, int):
            self.layer.SetNumProp(f'{self.ax}.type', value)
        elif isinstance(value, str):
            scales = {'linear': 1, 'log10': 2, 'probability': 3, 'probit': 4, 'reciprocal': 5,
                      'offset_reciprocal': 6, 'logit': 7, 'ln': 8, 'log2': 9}
            self.scale = scales.get(value)
        else:
            raise TypeError('unknown scale type')
        return self.scale

    @property
    def limits(self):
        """
        axis scale limits

        Parameters:

        Returns:
            (tuple) from, to, increment

        Examples:
            from, to, inc = ax.limits
        """
        return self.sfrom, self.sto, self.sstep

    @limits.setter
    def limits(self, limits):
        """
        Property setter returns the scale limits (from` and `to) for the axis.

        Parameters:
            limits (tuple): (from, to)

        Returns:
            None

        Examples:
            ax.limits = (1, 10, 1)
        """
        self.set_limits(limits)

    def set_limits(self, begin=None, end=None, step=None):
        """

        Sets scale from and to values for the axis

        Parameters:
            begin (float, tuple, list): The From value of limits, or specify [begin,end], or [begin,end,step]
            end (float): The To value of limits. None leaves the value unchanged.
            step (float): The Increment of the limits. None leaves the value unchanged.

        Returns:
            (tuple) from, to, increment

        Examples:
            ax.set_limits(step=0.3)
            ax.set_limits(5, 15)
        """
        #if auto:
            #cmd = {'x': 'yzm', 'y': 'xzm', 'z': 'xym'}[self.ax]
            #self.layer.LT_execute(f'layer -ar{cmd}e')
            #return self.limits
        #elif sto is None and isinstance(sfrom, (tuple, list)):
        #    sfrom, sto = sfrom
        if end is None and step is None and isinstance(begin, (tuple, list)):
            if len(begin) == 3:
                begin, end, step = begin
            elif len(begin) == 2:
                begin, end = begin
            else:
                raise ValueError('must specify 2 or 3 values')
            
        if begin is not None:
            self.sfrom = begin
        if end is not None:
            self.sto = end
        if step is not None:
            self.sstep = step
        return self.limits

    @property
    def title(self):
        """
        Property getter for axis title.

        Parameters:

        Returns:
            (str) Title

        Examples:
            txt = ax.title
        """
        label_name = {'x': 'xb', 'x2': 'xt', 'y': 'yl', 'y2': 'yr', 'z': 'zb', 'z2': 'zf'}[self.ax]
        label = GLayer(self.layer).label(label_name)
        if label is None:
            return ''
        return label.text

    @title.setter
    def title(self, value):
        """
        Property setter for axis title.

        Parameters:
            value (str): Title text

        Returns:
            (str) Title

        Examples:
            ax.title = 'X Axis'
        """
        label_name = {'x': 'xb', 'x2': 'xt', 'y': 'yl', 'y2': 'yr', 'z': 'zb', 'z2': 'zf'}[self.ax]
        cmd = f'label -{label_name} {value}'
        self.layer.LT_execute(cmd)
        return self.title


class Label(BaseObject):
    """
    This class represents an instance of a text object on a GLayer.
    """
    def __init__(self, obj, layer):
        self.layer=layer
        super().__init__(obj)
        
    def remove(self):
        """
        Deletes label.

        Parameters:

        Returns:
            None
        """
        self.obj.Destroy()
    '''
    def __repr__(self) -> str:
        if self.obj.IsValid():
            return f'Label named {self.name} in [{self.layer.GetParent().Name}]{self.layer.Name}'
        else:
            raise RuntimeError('label no longer exists')
    '''
    @property
    def color(self):
        """
        Property getter returns the RGB color of the text object as a tuple (Red, Green, Blue)

        Parameters:

        Returns:
            (tuple) r,g,b

        Examples:
            label = g[0].label('text')
            red, green, blue = label.color
        """
        orgb = self.get_int('color')
        return to_rgb(orgb)

    @color.setter
    def color(self, rgb):
        """
        Property setter for the RGB color of the text object

        Parameters:
            rgb(int, str, tuple): various way to specify color, see function ocolor(rgb) in op.utils

        Returns:
            None

        Examples:
            label = g[0].label('text')
            label.color = 'Red'
            label.color = 3            # 'Green'
            label.color = '#00f'       # 'blue'
            label.color = '#0000ff'    # 'blue'
            label.color = [0, 255, 0]  # 'green'
        """
        self.set_int('color', ocolor(rgb))
    @property
    def text(self) -> str:
        """
        Property getter for object text.

        Parameters:

        Returns:
            (str) Object text
        """
        return self.obj.Text

    @text.setter
    def text(self, text: str) -> str:
        """
        Property setter for object text.

        Parameters:
            value (str): Text

        Returns:
            (str) Object text
        """
        self.obj.Text = text
        return self.text

class Plot(BaseObject):
    """
    This class represents an instance of a data plot in a GLayer.
    """
    def __init__(self, obj, layer):
        self.layer=layer
        super().__init__(obj)

    @property
    def color(self):
        """
        returns the RGB color of the plot object as a tuple (Red, Green, Blue)

        Parameters:

        Returns:
            (tuple) r,g,b

        Examples:
            p = g[0].plot_list()[0]
            red, green, blue = p.color
        """
        nplot = self.index()+1
        orgb = int(self.layer.GetNumProp(f'plot{nplot}.color'))
        return to_rgb(orgb)

    @color.setter
    def color(self, rgb):
        """
        set for the RGB color of the plot object

        Parameters:
            rgb(int, str, tuple): various way to specify color, see function ocolor(rgb) in op.utils

        Returns:
            None

        Examples:
            p = g[0].add_plot(wks,'(0,1)')
            p.color = '#0f0'
        See Also:
            ocolor(rgb)
        """
        nplot = self.index()+1
        self.layer.SetNumProp(f'plot{nplot}.color', ocolor(rgb))
    def remove(self):
        """
        Deletes plot.

        Parameters:

        Returns:
            None
        """
        self.obj.Destroy()

    def change_data(self, wks, **kwargs):
        '''
        change the data source for an existing data plot
        
        Parameters:
            wks (WSheet): the worksheet to use
            kwargs : columns and the corresponding axis
            
        Examples:
            wks = op.find_sheet('w', 'Book1')
            gl = op.find_graph('Graph1')[0]
            dp = gl.plot_list()[0]
            dp.change_data(wks, x='C', y='D')
        '''
        data = []
        if wks:
            data = {desig: wks._find_col(obj) for desig, obj in kwargs.items()}
        else:
            data = kwargs
        for desig, obj in data.items():
            self.obj.ChangeData(obj, desig.upper())

class GLayer(BaseLayer):
    """
    This class represents an Origin Graph Layer, it holds an instance of a PyOrigin GraphLAyer
    """
    def rescale(self, skip = ''):
        """
        update limits to show all the data, including color scale range is colormap is used
        
        Parameters:
            skip(str): axis to skip(keep limits). Can be a combination of x,y,z,m where m is for colormap
        
        Returns:
            None
            
        Examples:
            gl=op.find_graph()[0]
            gl.rescale('x')#rescale Y only
            gl.rescale('m')#do not change colormap scale range
        """
        if len(skip) == 0:
            self.obj.LT_execute('layer -a')
        else:
            self.obj.LT_execute(f'layer -ar{skip}')
        
    def axis(self, ax):
        """
        Creates a new Axis object

        Parameters:
            ax (str): One of 'x', 'y', or 'z'

        Returns:
            (Axis)

        Examples:
            ax = lay.axis('x')
        """
        return Axis(self.obj, ax)

    @property
    def xlim(self):
        """
        Property getter for X axis limits.

        Parameters:

        Returns:
            New X axis limits
        """
        return self.axis('x').limits
    @xlim.setter
    def xlim(self, limits):
        """save as set_xlim"""
        return self.set_xlim(limits)

    @property
    def ylim(self):
        """
        See Also:
            xlim property getter
        """
        return self.axis('y').limits
    @ylim.setter
    def ylim(self, limits):
        """save as set_xlim"""
        return self.set_ylim(limits)

    @property
    def zlim(self):
        """
        See Also:
            xlim property getter
        """
        return self.axis('z').limits
    @zlim.setter
    def zlim(self, limits):
        """save as set_xlim"""
        return self.set_zlim(limits)
        
        
    def set_xlim(self, begin=None, end=None, step=None):
        """
        Sets X axis scale begin(From), end(To) and step(Increment)

        Parameters:
            begin (float or None): axis new From or Begin value.
            end (float or None): axis new To or End value.
            step (float or None): axis new Step or Increment value.

        Returns:
            New axis limits

        Examples:
            g = op.new_graph()
            g[0].set_xlim(0, 1)
            g[0].set_xlim(step=0.2)
        """
        return self.axis('x').set_limits(begin, end, step)

    def set_ylim(self, begin=None, end=None, step=None):
        """
        Sets Y axis scale begin(From), end(To) and step(Increment)

        Parameters:
            begin (float or None): axis new From or Begin value.
            end (float or None): axis new To or End value.
            step (float or None): axis new Step or Increment value.

        Returns:
            New axis limits

        Examples:
            g = op.new_graph()
            g[0].set_ylim(1, 100, 20)
            g[0].set_ylim(begin=0)
        """
        return self.axis('y').set_limits(begin, end, step)

    def set_zlim(self, begin=None, end=None, step=None):
        """
        Sets Z axis scale begin(From), end(To) and step(Increment)

        Parameters:
            begin (float or None): axis new From or Begin value.
            end (float or None): axis new To or End value.
            step (float or None): axis new Step or Increment value.

        Returns:
            New axis limits

        Examples:
            g = op.new_graph()
            g[0].set_zlim(0, 5)
            g[0].set_zlim(step=1)
        """
        return self.axis('z').set_limits(begin, end, step)


    @property
    def xscale(self):
        """
        Property getter for X axis scale type.

        Parameters:

        Returns:
            Scale type of axis

        Examples:
            st = lay.zscale
        """
        return self.axis('x').scale

    @xscale.setter
    def xscale(self, scaletype):
        """
        Property setter for X axis scale type.

        Parameters:
            scaletype (int or str): Supported string value of scaltype:
                ['linear', 'log10', 'probability', 'probit', 'reciprocal', 'offset_reciprocal', 'logit', 'ln', 'log2']

        Returns:
            New scale type of x axis

        Examples:
            lay.xscale = 'log10'
        """
        self.axis('x').scale = scaletype
        return self.xscale

    @property
    def yscale(self):
        """
        See Also:
            xscale property getter
        """
        return self.axis('y').scale

    @yscale.setter
    def yscale(self, scaletype):
        """
        See Also:
            xscale property setter
        """
        self.axis('y').scale = scaletype
        return self.yscale

    @property
    def zscale(self):
        """
         See Also:
            xscale property getter
        """
        return self.axis('z').scale

    @zscale.setter
    def zscale(self, scaletype):
        """
        See Also:
            xscale property setter
        """
        self.axis('z').scale = scaletype
        return self.axis('z').scale

    def label(self, name):
        """
        Get a Label instance by name.

        Parameters:
            name (str): name of the label to be attached
        Returns:
            (Label)

        Examples:
            g = op.new_graph()
            g[0].label('XB').remove()
        """
        lb = self.obj.GraphObjects(name)
        if lb is None:
            return None
        return Label(lb, self.obj)

    def remove_label(self, label):
        """
        Remove a label from a layer.

        Parameters:
            label (Label or str): Instance of Label or name of label to remove

        Returns:
            None

        Examples:
            g = op.new_graph()
            g[0].remove_label('xb') # g[0] is 1st layer.
        """
        if isinstance(label, Label):
            label.remove()
        elif isinstance(label, str):
            self.remove_label(self.label(label))
        else:
            raise TypeError('"label"" must ba an instance of either str or Label.')

    def remove(self, obj: any):
        if isinstance(obj, Label):
            self.remove_label(obj)
        elif isinstance(obj, Plot):
            self.remove_plot(obj)

    def plot_list(self):
        """
        Returns list of Plot objects in the layer.

        Parameters:

        Returns:
            (list)

        Examples:
            items = lay.plot_list()
            item = lay.plot_list()[0]
        """
        plist = []
        for dp in self.obj.DataPlots:
            plist.append(Plot(dp, self.obj))
        return plist

    def remove_plot(self, plot):
        """
        Removes plot from layer.

        Parameters:
            plot (Plot or int) Instance of Plot or index of plot to remove

        Returns:
            None

        Examples:
            lay.remove_plot(3)
        """
        if isinstance(plot, Plot):
            plot.remove()
        elif isinstance(plot, int):
            self.remove_plot(self.plot_list()[plot])
        else:
            raise TypeError('"plot"" must ba an instance of either Plot or int.')
    @staticmethod
    def _plot_type(type):
        if isinstance(type, int):
            return type
        if len(type) > 1:
            plttypes = {
                "line":'l',
                "scatter":'s',
                }
            type = plttypes[type]

        plottypes = {
            "l": 200,
            "s": 201,
            }
        return plottypes[type]
            
    def _add_plot(self, ranges, type):
        if isinstance(ranges, str):
            return self.obj.AddPlotFromString(ranges, self._plot_type(type))
        return self.obj.AddPlot(ranges, self._plot_type(type), True)
    
    def _to_lt_str(selt, col, wks):
        """return empty str if not valid, otherwise return LabTalk column index as str"""
        if isinstance(col,str):
            if len(col) == 0 or col == '#':
                return col
            temp = wks.lt_col_index(col)
            if temp < 1:
                return ''
            return str(temp)
        if not isinstance(col,int):
            raise ValueError('invalid column specification')
        if col < 0:
            return ''
        return str(col + 1)
        
    def add_plot(self, obj, coly=-1, colx=-1, colz=-1, type = 's'):
        """
        Add plot to graph layer
        Parameters:
            wks(WSheet): worksheet to plot from
            coly(int or str): Y column to plot
            colx(int or str): X column, it not specified, X column on the left, can use '#' to indicate use row number as X
            colz(int or str): Z column to plot
            type(str): 'l'(Line Plot) 's'(Scatter Plot)
        Returns:
            (Plot): the newly added Plot object

        Examples:
            gl = op.find_graph()[0]
            wks = op.find_sheet('w','Book1')
            p = gl.add_plot(wks,'C')
            p.color = '#5f0'
            gl.rescale()
            gl.add_plot(wks,1,3)#C(x), B(y)
            gl.add_plot(wks,1,'#')#B(y) vs Row Numbers
        """
        if isinstance(obj, origin_class('DataRange')) or isinstance(obj, str):
            return self._add_plot(obj, type)
        wks = obj
        colx = self._to_lt_str(colx, wks)
        coly = self._to_lt_str(coly, wks)
        colz = self._to_lt_str(colz, wks)
        if len(colz):
            if coly:
                colspec = f'({colx}, {coly}, {colz})'
            else:
                colspec = colz
        else:
            colspec = f'({colx}, {coly})'
        srange = wks.lt_range(False)
        dp = self._add_plot(f'{srange}!{colspec}', type)
        if dp is None:
            return None
        return Plot(dp, self.obj)

    def add_mplot(self, ms, z, x=-1, y=-1, cm=-1, type=103):
        '''
        plot a 3D parametric surface
        
        Parameters:
            ms (MSheet): matrix sheet with the data to plot
            z (int): matrix object index for Z
            x (int): matrix object index for X, must be > z
            y (int): matrix object index for Y, must be > z
            
        Examples:
            #assmues X,Y,Z contains the mesh grid data
            data=[]
            data.append(Z)
            data.append(X)
            data.append(Y)
            ms = op.new_sheet('m')
            ms.from_np(np.array(data))
            gl = op.new_graph(template='GLparafunc')[0]
            gl.add_mplot(ms, 0, 1, 2)
        '''
        self.add_plot(ms, colz=z, type=type)

        mo = ms.obj[z]
        moname = mo.DatasetName
        x = 0 if x < 0 else x + 1
        y = 0 if y < 0 else y + 1
        self.obj.LT_execute(f'set {moname} -zx {x};set {moname} -zy {y};')

        colormapname = ''
        if cm >= 0:
            mcm = ms.obj[cm]
            colormapname = mcm.DatasetName
            self.obj.LT_execute(f'set {moname} -b3c {colormapname};')

        
class GPage(BasePage):
    """
    This class represent an Origin Graph Page, it holds an instance of a PyOrigin GraphPage
    """
    def __repr__(self):
        return 'GPage: ' + self.obj.GetName()

    def __getitem__(self, index):
        return GLayer(self.obj.Layers(index))

    def __iter__(self):
        for elem in self.obj.Layers():
            yield GLayer(elem)

    def save_fig(self, path='', type='auto', replace=True, width = 0):
        """
        export to a file

        Parameters:
        -------
        path(str): if not specified, then to UFF, or if graph has theme, then last exported location in that graph.
                    If specifed, can include file name with extention, which will then control image type
        type(str): if auto, then use graph last exported settings, otherwise 'png', 'emf' etc
        replace(bool): if file existed, silently replace or not
        width(int): Pixcels. 0 will use last exported setting only if type=auto, if not auto, better specify this.

        Return:
        -------
            if success, returns the fullpath file name generated, otherwise an empty string

        Examples
        -------
        >>> g = op.find_graph('graph1')
        >>> g.save_fig(op.path() + 'test1.png', width=500)
        >>> g.save_fig()#use last GUI export graph settings remembered inside graph, which may include path

        """
        fname=''
        if len(path):
            path0 = path
            path, fname, ext=get_file_parts(path)
            if len(ext):
                if type!='auto' and type != ext[1:]:
                    raise ValueError('must not specify type and different from file extension')
                type = ext[1:]
            else:
                path = path0#d:\\test\\test, need to restore
                path = last_backslash(path,'t')#expgraph has trouble with path ending with \\
                fname=''
        exp = 'expgraph -sw'#silent warning messages
        if type=='auto':
            exp += ' -t <book>'
        else:
            exp += ' type:=' + str(type)

        if len(fname):
            exp += f' filename:="{fname}"'
        if replace:
            exp += ' overwrite:=replace'
        else:
            exp += ' overwrite:=skip'
        if width > 0:
            exp += f' tr1.Unit:=2 tr1.Width:={width}'
        if len(path):
            exp += f' path:="{path}"'
        self.obj.LT_execute(exp)
        return po.LT_get_str('__LASTEXP')

    def add_layer(self, type = 0):
        """
        Adds a new layer to GPage. Type is based on types specified here:https://www.originlab.com/doc/X-Function/ref/layadd
        but should use the index equivalent of listed types (e.g. `righty` would be 2)

        Parameters:
            type (int): Type of layer to add. See description above foir how to deduce

        Returns:
            (GLayer): the newly added graph layer

        Examples:
            g = op.find_graph('graph1')
            gl=g.add_layer(2) # righty
            gl.xlim=[-2,1]
        """
        lt = 'layadd type:=' + str(type)
        self.obj.LT_execute(lt)
        nlayers = self.obj.Layers.GetCount()
        return GLayer(self.obj.Layers(nlayers-1))



